''''''''''''''''''''''''''''''''''''''''''''
''  Sand Demo
''  Written 2007 by Dennis Ferron
''
''''''''''''''''''''''''''''''''''''''''''''
CON

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  x_tiles = 16
  y_tiles = 12

  paramcount = 14       

CON
                                

  ScreenH = 192
  ScreenW = 256

  TileH = 16
  TileW = 16

  ScreenBytes   = ScreenW * ScreenH / 4 + 64
  LineBytes     = 256/4         ' 256 x-resolution, 2 bits per pixel, 4 pixels per byte.

  ' The amount of space to skip forward to reach the next strip of tiles
  ' in the pixel buffer.
  TileColSkip = TileH * y_tiles

  Red = $38
  Blue = $98
  Cyan = $4E
  Peach = $AE
  DarkGreen = $6B
  DarkBrown = $BA

  Black = $02
  DarkGrey = $03
  MediumGrey = $04
  Grey = $05
  LightGrey = $06
  White = $07

  OverscanColor = Cyan
  BackColor = DarkGrey
  SandColor = Peach
  EnemyColor = Red
  AllyColor = Blue

  BackPixel = 0
  SandPixel = 1
  EnemyPixel = 2
  AllyPixel = 3

  InitSandH = (3 * ScreenH) / 4
  InitSandW = (1 * ScreenW) / 4

  TowerHeight = 2

VAR

  byte  display[ScreenBytes]

  long  tv_status     '0/1/2 = off/visible/invisible           read-only
  long  tv_enable     '0/? = off/on                            write-only
  long  tv_pins       '%ppmmm = pins                           write-only
  long  tv_mode       '%ccinp = chroma,interlace,ntsc/pal,swap write-only
  long  tv_screen     'pointer to screen (words)               write-only
  long  tv_colors     'pointer to colors (longs)               write-only               
  long  tv_hc         'horizontal cells                        write-only
  long  tv_vc         'vertical cells                          write-only
  long  tv_hx         'horizontal cell expansion               write-only
  long  tv_vx         'vertical cell expansion                 write-only
  long  tv_ho         'horizontal offset                       write-only
  long  tv_vo         'vertical offset                         write-only
  long  tv_broadcast  'broadcast frequency (Hz)                write-only
  long  tv_auralcog   'aural fm cog                            write-only

  long  colors[2]

OBJ

  tv    : "raster-tv"
  phys1 : "sand_physics"
  phys2 : "sand_physics"
  phys3 : "sand_physics"
  phys4 : "sand_physics"
  phys5 : "sand_physics"
  phys6 : "sand_physics"

PRI MakePalEntry(Color0, Color1, Color2, Color3) : Pal

  Pal := Color0 + (Color1 << 8) + (Color2 << 16) + (Color3 << 24)  

PUB start | i, j, c, dx, gx, gy, landed

  'start tv
  longmove(@tv_status, @tvparams, paramcount)
  tv_screen := @display
  tv_colors := @colors
  tv.start(@tv_status)

  'init colors
  'Color 0 is black ($02)
  'Color 1 is sandy ($AE)
  'Color 2 is red   ($38)
  'Color 3 is blue  ($98)
  colors[1] :=  MakePalEntry(BackColor, SandColor, EnemyColor, AllyColor)  ' $9838AE02

  ' set overscan color
  colors[0] := MakePalEntry(OverscanColor, BackColor, BackColor, BackColor) ' $0202024E

  phys1.start(@display)
  waitcnt(1000_000 + cnt)
  phys2.start(@display)
  waitcnt(1000_000 + cnt)
  phys3.start(@display)
  waitcnt(1000_000 + cnt)
  phys4.start(@display)
  waitcnt(1000_000 + cnt)
  phys5.start(@display)
  waitcnt(1000_000 + cnt)
  phys6.start(@display)
  
  repeat 
    gx := c & $FF 
    gy := 10
    c := ?c

    SetPixel(gx, gy, 1)
    waitcnt(500_000 + cnt)    
    SetPixel(10, gy, 1)
    waitcnt(700_000 + cnt)        
    SetPixel(160, gy, 1)
    SetPixel(130, gy, 1)
    waitcnt(300_000 + cnt)

    ' Create a barrier
    repeat gx from 80 to 100
      repeat gy from 100 to 102
        SetPixel(gx, gy, 2)

    ' Create a sink-hole
    repeat gx from 200 to 207
      repeat gy from 180 to 188
        SetPixel(gx, gy, 2)


PRI DoSandPhysics | sandx, sandy, movex, movey, towery

  repeat sandy from ScreenH - 1 to ScreenH - 30
    repeat sandx from ScreenW / 2 - 10 to ScreenW / 2 + 10
      movex := 0
      movey := 0
      if IsSand(sandx, sandy)
        if sandy < ScreenH - 1 and not IsSand(sandx, sandy + 1)
          movey := 1
        elseif SandTower(sandx, sandy)
          if not IsSand(sandx + 1, sandy)
            movex := 1
          elseif not IsSand(sandx - 1, sandy)
            movex := -1

        SetPixel(sandx, sandy, BackPixel)
        sandx += movex
        sandy += movey
        SetPixel(sandx, sandy, SandPixel)

PRI SandTower(px, py) | i

  repeat i from 0 to TowerHeight - 1
    if not IsSand(px, py - i)
      return false

  return true          
          
PRI IsSand(px, py) : is_sand

  is_sand := (GetPixel(px, py) == SandPixel)

PRI GetPixel(px, py) : pixel_color 

  pixel_color := display[(px >> 2) + (py * LineBytes)]
  pixel_color >>= ((px & %11) << 1)
  pixel_color &= %11

PRI SetPixel(px, py, pixel_color) | index, old_color, shift_amount

  index := (px >> 2) + (py * LineBytes)
  old_color := display[index]

  shift_amount := (px & %11) << 1

  old_color &= !(%11 << shift_amount)
  old_color |= (pixel_color << shift_amount)

  display[index] := old_color

PRI GetPixelShift(px) : shift_amount

  shift_amount := (px & %11) << 1

PRI GetPixelIndex(px, py) : pixel_index

  pixel_index := (px >> 2) + (py * LineBytes)

DAT

tvparams                long    0               'status
                        long    1               'enable
                        long    %010_0101       '%001_0101       'pins
                        long    %0001           'mode
                        long    0               'screen
                        long    0               'colors
                        long    x_tiles         'hc
                        long    y_tiles         'vc
                        long    10              'hx
                        long    1               'vx
                        long    0               'ho
                        long    0               'vo
                        long    0               'broadcast
                        long    0               'auralcog