'' Physics engine for making sand collapse and flow.
''
CON

  ScreenH = 192
  ScreenW = 256

  BackPixel = 0
  SandPixel = 1
  EnemyPixel = 2
  AllyPixel = 3

  TowerHeight = 2

VAR

  long  cog


PUB start(display_ptr) : okay

'' Start Sand Physics - starts a cog
'' returns false if no cog available

  stop
  okay := cog := cognew(@entry, (display_ptr)) + 1


PUB stop

'' Stop sand physics - frees a cog

  if cog
    cogstop(cog~ - 1)


DAT

'*****************************************
'* Assembly language Sand Physics Engine *
'*****************************************

                        org
'
'
' Entry
'
entry                   mov     screen,par              ' Get display pointer

                        mov     pixelptr, screen        ' DLF:  screen points directly at the buffer,
                                                        ' and we scan it sequentially.'
                        mov x_dir, neg1
:scan_screen
                        ' Multiple cogs more likely to "collide" if going
                        ' opposite directions; so this is temporarily disabled.
              '          cmps x_dir, #1 wz
              'if_e      mov x_dir, neg1
              'if_ne     mov x_dir, #1
              
                        mov sandy, #(ScreenH - 1)
:loop_y
                        mov x_count, #(ScreenW)                                

                        cmps x_dir, #1 wz
              if_e      mov sandx, #0  
              if_ne     mov sandx, #(ScreenW - 1)

              
:loop_x
                        mov movex, #0
                        mov movey, #0

                        mov px, sandx
                        mov py, sandy
                        call #IsSand
                        cmp is_sand, T wz
              if_ne     jmp #:not_sand

                        cmp sandy, #(ScreenH - 10)  wc, wz
              if_a      jmp #:not_falling
                        
                        mov px, sandx
                        mov py, sandy
                        add py, #1
                        call #IsSand
                        cmp is_sand, T wz
              if_e      jmp #:not_falling

                        mov movey, #1
                        jmp #:move_sand

:not_falling
                        
                        mov px, sandx
                        mov py, sandy
                        call #IsSandTower
                        cmp is_tower, T wz
              if_ne     jmp #:not_tower

                        cmps x_dir, neg1 wz
              if_ne     jmp #:try_right
              if_e      jmp #:try_left

:left_failed
                        cmps x_dir, neg1 wz
              if_e      jmp #:try_right
              if_ne     jmp #:nevermind

:right_failed
                        cmps x_dir, neg1 wz
              if_ne      jmp #:try_left
              if_e     jmp #:nevermind

:try_left
                       ' Left side of tower
                        mov px, sandx
                        mov py, sandy
                        sub px, #1                
                        cmps zero, px wz
              if_e      jmp #:left_failed      
                        call #IsSand
                        cmp is_sand, T wz
              if_e      jmp #:left_failed

                        neg movex, #1
                        jmp #:move_sand
                        
:try_right
                        mov px, sandx
                        mov py, sandy
                        add px, #1
                        cmps px, #ScreenW - 1 wc
              if_nc     jmp #:right_failed
                        call #IsSand
                        cmp is_sand, T wz
              if_e      jmp #:right_failed
              
                        mov movex, #1
                        jmp #:move_sand

:move_sand
           
                        mov px, sandx
                        mov py, sandy
                        mov pixelcolor, #3
                        call #SetPixel

                        mov px, sandx
                        mov py, sandy
                        add px, movex
                        add py, movey
                        mov pixelcolor, #1
                        call #SetPixel   
:nevermind 
:not_tower
:not_sand
                        add sandx, x_dir
                        djnz    x_count, #:loop_x
                        'djnz    sandx, #:loop_x
                        djnz    sandy, #:loop_y
                        jmp     #:scan_screen


                        
IsSandTower
                        mov is_tower, T
                        mov tempy, py
                        mov count, #TowerHeight             
:loop
                        call #IsSand
                        cmp is_sand, T wz
              if_ne     mov is_tower, Nil
              if_ne     jmp #:not_sand
                        sub py, #1
                        djnz count, #:loop
:not_sand
                        mov py, tempy
IsSandTower_ret         ret
 
                        
                                               

IsSand
                        call #GetPixel
                        cmp pixelcolor, #(SandPixel) wz
              if_e      mov is_sand, T
              if_ne     mov is_sand, Nil
IsSand_ret              ret



GetPixel
                        mov pixelcolor, #SandPixel
                        call #GetPixelPtr               ' Find where to get the pixel data from
                        rdbyte pixelcolor, pixelptr     ' Get the pixels out of RAM
                        call #GetPixelShift             ' Find how far to shift the bits for the pixel we want.
                        shr pixelcolor, pixelshift      ' Bring the pixel bits to the bottom.
                        and pixelcolor, #3              ' Strip out all but the 2 bits we are interested in.
GetPixel_ret            ret



SetPixel
                        call #GetPixelPtr               ' Find where to get the pixel data from
                        rdbyte oldcolor, pixelptr       ' Get the old pixels out of RAM
                        mov newcolor, pixelcolor        ' Xfer pixelcolor to working register so we don't overwrite it.
                        mov mask, #3                    ' Set up the mask with two bits set.
                        and newcolor, mask              ' Make sure we set only 2 bits with the new color.
                        call #GetPixelShift             ' Find out how much to shift to get the pixel we want.
                        shl newcolor, pixelshift        ' Place the new color bits in the right locations.
                        shl mask, pixelshift            ' Place the mask bits in the right locations.
                        xor mask, #$FF                  ' Invert the mask.
                        and oldcolor, mask              ' Clear the old pixel bits.
                        or oldcolor, newcolor           ' Set the new pixel bits.
                        wrbyte oldcolor, pixelptr       ' Return the combined new and old pixels.
SetPixel_ret            ret
                                      
                        
GetPixelPtr
                        max px, #(ScreenW-1)
                        max py, #(ScreenH-1)
                        mov pixelptr, px            ' Don't modify px.  Use return value as workspace.
                        shr pixelptr, #2            ' Find x offset first; 4 pixels per byte.

                        mov tempy, py                   ' Don't modify py.  Use tempy as workspace.
                        shl tempy, #6                   ' Find line offset; 64 bytes per line                                        
                         
                        add pixelptr, tempy             ' Combine x offset with line offset to get real offset.
                        add pixelptr, screen            ' Combine offset with location of the screen buffer.

GetPixelPtr_ret         ret                             ' pixelpointer = address of pixel at (px, py)
                        

GetPixelShift
                        mov pixelshift, px              ' Don't modify px.  Use return value as workspace.
                        and pixelshift, #3              ' Find number of pixels to shift.
                        shl pixelshift, #1              ' Two bits per pixel, so doubled to get bits to shift.
GetPixelShift_ret       ret



zero    long  0
neg1    long  -1
Nil     long  0
T       long  -1


'
'
' Uninitialized data
'

' DLF
move_tower_fail         res     1
move_tower_pass         res     1
x_count                 res     1
x_dir                   res     1
is_sand                 res     1
is_tower                res     1
pixelptr                res     1
sandx                   res     1                       'display
sandy                   res     1
movex                   res     1
movey                   res     1
px                      res     1
py                      res     1
tempx                   res     1
tempy                   res     1
screen                  res     1
pixelcolor              res     1
oldcolor                res     1
newcolor                res     1
mask                    res     1
pixelshift              res     1
count                   res     1